#include <string.h>

#include "as31glue.h"
#include "printf.h"
#include "paulmon2.h"
#include "display.h"
#include "malloc.h"
#include "dirlist.h"
#include "playlist.h"
#include "stricmp.h"
#include "rand.h"
#include "main.h"

simm_id current_playlist;
simm_id first_playlist;
simm_id last_playlist;
simm_id start_rand_playlist;
simm_id head_rand_playlist;
simm_id tail_rand_playlist;
simm_id start_all_rand;
simm_id head_all_rand;
simm_id tail_all_rand;
unsigned int num_unrand_lists;
unsigned int num_unrand_all_files;

static playlist_list_t *dirrand_current_list;
static unsigned int addr6_backup, addr7_backup;


play_modes_t play_mode;


static void dirrand_initialize(void);
static simm_id dirrand_new_list(simm_id /* playlist_list_t */ list_id, unsigned char use_current_file);
static void dirrand_randomize_incremental(simm_id /* playlist_t */ current);
static void plstrand_initialize(void);
static void plstrand_randomize_incremental(simm_id /* playlist_list_t */ list_id);
static void allrand_initialize(void);
static void allrand_randomize_incremental(simm_id /* playlist_list_t */ list_id);


// #define VERBOSE_DEBUG


void playlist_init(void)
{
  first_playlist=0;
  last_playlist=0;
  current_playlist=0;
}


extern data simm_id temp_id;


void playlist_create(simm_id pl_name) 
{
  // simm_id temp_id;
  playlist_list_t * pl_l_struct;

  temp_id = simm_malloc((unsigned int)(sizeof(struct playlist_list_struct)));

  if(first_playlist == 0)
    {
      first_playlist = temp_id;
      last_playlist = temp_id;
      pl_l_struct = addr6(temp_id);
      pl_l_struct->prev_sequential = 0;
    }
  else
    {
      pl_l_struct = Addr6(last_playlist);
      pl_l_struct->next_sequential = temp_id;
      pl_l_struct = addr6(temp_id);
      pl_l_struct->prev_sequential = last_playlist;
      last_playlist = temp_id;
    }
  current_playlist = temp_id;
  pl_l_struct->next_sequential = 0;
  pl_l_struct->playlist_start_sequential = 0;
  pl_l_struct->playlist_end_sequential = 0;
  pl_l_struct->num_unrand_files = 0;
  pl_l_struct->name = pl_name;
  //printf("playlist %s created\r\n", Addr6(pl_name));
}




void playlist_add_file(simm_id /* struct filelist_struct */ fl_rec) // Add file to current playlist
{
  // xdata simm_id temp_id;
  playlist_list_t * pl_l_struct;
  playlist_t * pl_struct;
  filelist_t * fl_struct;
  
  pl_l_struct = Addr6(current_playlist);
  pl_l_struct->num_unrand_files++;

  temp_id = simm_malloc((unsigned int)(sizeof(struct playlist_struct)));
  if (pl_l_struct->playlist_start_sequential == 0)
    {
      pl_l_struct->playlist_start_sequential = temp_id;
      pl_l_struct->playlist_end_sequential = temp_id;
      pl_struct = addr7(temp_id);
      memset(pl_struct, 0, sizeof(struct playlist_struct));
      pl_struct->prev_sequential = 0;  // done by memset
    }
  else
    {
      pl_struct = Addr7(pl_l_struct->playlist_end_sequential);
      pl_struct->next_sequential = temp_id; 
      pl_struct = addr7(temp_id);
      memset(pl_struct, 0, sizeof(struct playlist_struct));
      pl_struct->prev_sequential = pl_l_struct->playlist_end_sequential;
      pl_l_struct->playlist_end_sequential = temp_id;
    }
  pl_struct->playlist = current_playlist;
  pl_struct->next_sequential = 0;  // done by memset
  pl_struct->fl_rec = fl_rec;
  fl_struct = Addr6(fl_rec);
  fl_struct->pl_entry = temp_id;
}


void playlist_show(void) // Shows the Current Playlist
{
	playlist_t * pl_struct;
	filelist_t * fl_struct;
  
	temp_id = ((playlist_list_t *)Addr7(current_playlist))->playlist_start_sequential;
	while(temp_id != 0) {
		pl_struct = addr6(temp_id);
		fl_struct = Addr7(pl_struct->fl_rec);
		print_hex32(pl_struct->prev_sequential);
		print_hex32(temp_id);
		print_hex32(pl_struct->next_sequential);
		print_str(Addr6(fl_struct->long_name));
		print_crlf();
		temp_id = ((playlist_t *)addr6(temp_id))->next_sequential;
	}
}


void playlist_delete(playlist_list_t *p)
{
	print("deleting empty playlist: ");
	print_str(Addr7(p->name));
	print_crlf();
	if (p->prev_sequential) {
		((playlist_list_t *)Addr7(p->prev_sequential))->next_sequential =
			p->next_sequential;
	} else {
		first_playlist = p->next_sequential;
	}
	if (p->next_sequential) {
		((playlist_list_t *)Addr7(p->next_sequential))->prev_sequential =
			p->prev_sequential;
	} else {
		last_playlist = p->next_sequential;
	}
	simm_free(simm_id_from_pointer(p));
}

void playlist_purge_empty(void)
{
	playlist_list_t * pl_l_struct;
	
	temp_id = first_playlist;
	while (temp_id) {
		pl_l_struct = addr6(temp_id);
		temp_id = pl_l_struct->next_sequential;
		if (pl_l_struct->playlist_start_sequential == 0) {
			playlist_delete(pl_l_struct);
		}
	}
}

void playlist_show_all() // Show all playlists
{
	playlist_list_t * pl_l_struct;

	current_playlist = first_playlist;
	while (current_playlist != 0) {
		pl_l_struct = Addr6(current_playlist);
		print("\r\nPlaylist: ");
		print_hex32(pl_l_struct->prev_sequential);
		print_hex32(current_playlist);
		print_hex32(pl_l_struct->next_sequential);
		print("\"");
		print_str(Addr7(pl_l_struct->name));
		print("\"\r\n");
		playlist_show();
		print_crlf();
		pl_l_struct = Addr6(current_playlist);
		current_playlist = pl_l_struct->next_sequential;
	}
	current_playlist = first_playlist; // Reset to first
}







void play_mode_change(void)
{
	switch (play_mode) {
	case MODE_DIRONLY:    // sequential within one directory
	case MODE_SEQNTIAL:   // sequential among all directories
		break;
	case MODE_PLSTRAND:   // seq within each dir, choose dirs randomly
		plstrand_initialize();
		update_rand_seed_needed = 1;
		break;
	case MODE_DIRRAND:    // randomly choose files within 1 dir
		dirrand_initialize();
		update_rand_seed_needed = 1;
		break;
	case MODE_ALLRAND:    // randomly choose files among all dirs
	case MODE_KEYRAND:    // play all files sequentially until a key is pressed
		allrand_initialize();
		update_rand_seed_needed = 1;
		break;
	default:
		break;
	}
	update_mode_needed = 1;
	update_playlist_state_needed = 1;
	param_restart_write_timer(15);
}


void save_play_mode(void)
{
	ibuf[0] = play_mode;
	write_flash_param(PARAM_PLAY_MODE, ibuf);
	update_mode_needed = 0;
	print("saved play mode = ");
	print_hex16(play_mode);
	print_crlf();
}


void play_mode_toggle(void)
{
	if (play_mode == MODE_SEQNTIAL) {
		play_mode = MODE_ALLRAND;
	} else {
		play_mode = MODE_SEQNTIAL;
	}
	play_mode_change();
}

void play_mode_next(void)
{
	printf("play_mode = %d\r\n", play_mode);
	if (play_mode >= MODE_MAX) {
		play_mode = MODE_MIN;
	} else {
		play_mode++;
	}
	printf("play_mode = %d\r\n", play_mode);
	play_mode_change();
}

void play_mode_prev(void)
{
	if (play_mode <= MODE_MIN) {
		play_mode = MODE_MAX;
	} else {
		play_mode--;
	}
	play_mode_change();
}

void play_mode_set(play_modes_t new_play_mode)
{
	play_mode = new_play_mode;
	play_mode_change();
}











#pragma CALLEE-SAVES sdcc_bug_587536_workaround
static void sdcc_bug_587536_workaround(void xdata *p)
{ p; }

/*
simm_id file_ptr(playlist_t *p, xdata int file_num)
{
	int i;
	playlist_t 
	for(i=file_num;i!=0;i++) {
		
	}
	
}
*/

simm_id next_file_ptr(playlist_t *p, unsigned char keyflag)
{
	#ifdef VERBOSE_DEBUG
	print("begin next_file_ptr, p=");
	print_hex32(simm_id_from_pointer(p));
	print_crlf();
	#endif

	switch (play_mode) {
	  case MODE_DIRONLY:	// sequential within one directory
		if (p->next_sequential) {
			return p->next_sequential;
		} else {
			return ((playlist_list_t *)Addr6(p->playlist))->playlist_start_sequential;
		}
	  case MODE_SEQNTIAL:	// sequential among all directories
	  case MODE_PLSTRAND:	// seq within each dir, choose dirs randomly
		if (p->next_sequential) {
			return p->next_sequential;
		} else {
			return next_playlist_ptr(p, 0);
		}
	  case MODE_DIRRAND:	// randomly choose files within 1 dir
		if (p->next_random == 0) {
			dirrand_randomize_incremental(simm_id_from_pointer(p));
		}
		sdcc_bug_587536_workaround(p);
		return p->next_random;
	  case MODE_KEYRAND:    // play all files sequentially until a key is pressed
		if (keyflag == 0) {
			// no key pressed so act as MODE_SEQNTIAL
			if (p->next_sequential) {
				return p->next_sequential;
			} else {
				return next_playlist_ptr(p, 0);
			}
		}
		sdcc_bug_587536_workaround(p);
	    // else fall through
	  case MODE_ALLRAND:	// randomly choose files among all dirs
		if (p->next_random == 0) {
			allrand_randomize_incremental(simm_id_from_pointer(p));
		}
		sdcc_bug_587536_workaround(p);
		return p->next_random;
	  default:
		break;
	}
	return 0;
}

simm_id prev_file_ptr(playlist_t *p)
{

	switch (play_mode) {
	  case MODE_DIRONLY:	// sequential within one directory
		if (p->prev_sequential) {
			return p->prev_sequential;
		} else {
			return ((playlist_list_t *)Addr6(p->playlist))->playlist_end_sequential;
		}
	  case MODE_SEQNTIAL:	// sequential among all directories
	  case MODE_PLSTRAND:	// seq within each dir, choose dirs randomly
		if (p->prev_sequential) {
			return p->prev_sequential;
		} else {
			p = (playlist_t *)addr6(prev_playlist_ptr(p));
			return ((playlist_list_t *)Addr6(p->playlist))->playlist_end_sequential;
		}
	  case MODE_DIRRAND:	// randomly choose files within 1 dir
		if (p->prev_random == 0) {
			dirrand_randomize_incremental(simm_id_from_pointer(p));
		}
		sdcc_bug_587536_workaround(p);
		return p->prev_random;
	  case MODE_KEYRAND:    // play all files sequentially until a key is pressed
                                // no decision, must be here due to keypress 
	  case MODE_ALLRAND:	// randomly choose files among all dirs
		if (p->prev_random == 0) {
			allrand_randomize_incremental(simm_id_from_pointer(p));
		}
		sdcc_bug_587536_workaround(p);
		return p->prev_random;
	  default:
		break;
	}
	return 0;
}

simm_id next_playlist_ptr(playlist_t *p, unsigned char keyflag)
{
	playlist_list_t *list;

	list = (playlist_list_t *)Addr6(p->playlist);

	switch (play_mode) {
	  case MODE_KEYRAND:    // play all files sequentially until a key is pressed
		if (keyflag != 0) {
			// keypress so act as MODE_ALLRAND
			return next_file_ptr(p, 1);
		}
	    // else fall through
	  case MODE_DIRONLY:	// sequential within one directory
	  case MODE_SEQNTIAL:	// sequential among all directories
		if (list->next_sequential) {
			return ((playlist_list_t *)addr6(list->next_sequential))->playlist_start_sequential;
		} else {
			return ((playlist_list_t *)Addr6(first_playlist))->playlist_start_sequential;
		}

	  case MODE_DIRRAND:	// randomly choose files within 1 dir
		if (list->next_sequential) {
			return dirrand_new_list(list->next_sequential, 0);
		} else {
			return dirrand_new_list(first_playlist, 0);
		}
	  case MODE_PLSTRAND:	// seq within each dir, choose dirs randomly
		if (list->next_random == 0) {
			plstrand_randomize_incremental(simm_id_from_pointer(list));
		}
		sdcc_bug_587536_workaround(list);
		return ((playlist_list_t *)Addr6(list->next_random))->playlist_start_sequential;
	  case MODE_ALLRAND:	// randomly choose files among all dirs
		return next_file_ptr(p, 0);

	  default:
		break;
	}
	return 0;
}

simm_id prev_playlist_ptr(playlist_t *p)
{
	playlist_list_t *list;

	list = Addr6(p->playlist);

	switch (play_mode) {
	  case MODE_DIRONLY:	// sequential within one directory
	  case MODE_SEQNTIAL:	// sequential among all directories
		if (list->prev_sequential) {
			return ((playlist_list_t *)addr6(list->prev_sequential))->playlist_start_sequential;
		} else {
			return ((playlist_list_t *)Addr6(last_playlist))->playlist_start_sequential;
		}

	  case MODE_DIRRAND:	// randomly choose files within 1 dir
		if (list->prev_sequential) {
			return dirrand_new_list(list->prev_sequential, 0);
		} else {
			return dirrand_new_list(last_playlist, 0);
		}
	  case MODE_PLSTRAND:	// seq within each dir, choose dirs randomly
		if (list->prev_random == 0) {
			plstrand_randomize_incremental(simm_id_from_pointer(list));
		}
		sdcc_bug_587536_workaround(list);
		return ((playlist_list_t *)Addr6(list->prev_random))->playlist_start_sequential;

	  case MODE_KEYRAND:    // play all files sequentially until a key is pressed
                                // no decision, must be here due to keypress 
	  case MODE_ALLRAND:	// randomly choose files among all dirs
		return prev_file_ptr(p);

	  default:
		break;
	}
	return 0;
}







unsigned long
playlist_to_state(playlist_t *p)
{
	playlist_state_t state;
	playlist_list_t *list;

	if (play_mode == MODE_SEQNTIAL || play_mode == MODE_DIRONLY) {
		state.part.song_offset = 0;
		while (p->prev_sequential) {
			p = Addr7(p->prev_sequential);
			state.part.song_offset++;
		}
	} else {
		state.part.song_offset = 0xFFFF;
	}
	if (play_mode == MODE_SEQNTIAL || play_mode == MODE_DIRONLY
	  || play_mode == MODE_DIRRAND) {
		state.part.list_offset = 0;
		list = Addr7(p->playlist);
		while (list->prev_sequential) {
			list = Addr7(list->prev_sequential);
			state.part.list_offset++;
		}
	} else {
		state.part.list_offset = 0xFFFF;
	}
	print("playlist state = ");
	print_hex32(state.whole);
	print_crlf();
	return state.whole;
}


simm_id
state_to_playlist(unsigned long state_in)
{
	playlist_state_t state;
	playlist_t *song;
	playlist_list_t *list;

	state.whole = state_in;
	list = Addr7(first_playlist);
	if (state.part.list_offset != 0xFFFF) {
		while (state.part.list_offset && list->next_sequential) {
			list = Addr7(list->next_sequential);
			state.part.list_offset--;
		}
	}
	song = Addr7(list->playlist_start_sequential);
	if (state.part.song_offset != 0xFFFF) {
		while (state.part.song_offset && song->next_sequential) {
			song = Addr7(song->next_sequential);
			state.part.song_offset--;
		}
	}
	return simm_id_from_pointer(song);
}








/*************************************************************************************/
/* Randomization code for MODE_DIRRAND (randomly choose files within 1 dir)          */
/*************************************************************************************/

// Dir Rand mode is tricky to implement.  Though the purpose is to
// play all files in one directory in random order, the user can
// press next/prev direcory buttons to jump between directories.
// The next_playlist_ptr() and prev_playlist_ptr() functions need
// to report would 'would' happen, so it is not acceptable to store
// a single static variable of the "current_list" as was previously
// done.  Each directory needs to store its own randomized order
// so we can move between them and have all the buttons function
// in the way a user expects (songs come up in a seemingly very
// "random" order, but all next/prev buttons have highly predictable
// behavior when traversing to previously heard songs).  Initialization
// is also tricky, because the when the mode is initialized we are
// required to start on "current_file", but when a new directory is
// entered we're supposed to start at some "random" point in it.


static void
dirrand_show_current_playlist(void) // Shows the Current Playlist (randomized)
{
	playlist_t * pl_struct;
 
	temp_id = ((playlist_list_t *)Addr6(current_playlist))->playlist_start_random;
	pl_struct = (playlist_t *)addr6(temp_id);
	while (1) {
		print_hex32(simm_id_from_pointer(pl_struct));
		print_hex32(pl_struct->prev_random);
		print_hex32(pl_struct->next_random);
		print_hex32(temp_id);
		print_str((xdata char *)Addr7(((filelist_t *)Addr7(pl_struct->fl_rec))->long_name));
		print_crlf();
		if (pl_struct->next_random == temp_id) break;
		pl_struct = (playlist_t *)Addr6(pl_struct->next_random);
	}
	print_crlf();
}



static playlist_t *
dirrand_choose_file(void)
{
	unsigned int count;
	playlist_t *entry;
	//unsigned int r;

	entry = Addr7(dirrand_current_list->playlist_start_sequential);
	 //r = rand16();
	 //print("rand = ");
	 //print_hex16(r);
	 //print(", num_unrand_files = ");
	 //print_hex16(dirrand_current_list->num_unrand_files);
	 //print_crlf();
	count = rand16() % dirrand_current_list->num_unrand_files;
	#ifdef VERBOSE_DEBUG
	printf("begin dirrand_choose_file, count=%u, num_unrand=%u\r\n",
		count, dirrand_current_list->num_unrand_files);
	#endif
	while (1) {
		if (!entry->is_randomized) {
			#ifdef VERBOSE_DEBUG
			printf(" [%lx]", simm_id_from_pointer(entry));
			#endif
			if (count == 0) {
				#ifdef VERBOSE_DEBUG
				print("\r\nend dirrand_choose_file\r\n");
				#endif
				return entry;
			}
			count--;
		} else {
			#ifdef VERBOSE_DEBUG
			printf(" (%lx)", simm_id_from_pointer(entry));
			#endif
		}
		entry = Addr7(entry->next_sequential);
	}
}


static void
dirrand_initialize(void)
{
	temp_id = first_playlist;

	// mark all the directories as not randomized
	while (temp_id) {
		dirrand_current_list = (playlist_list_t *)addr6(temp_id);
		dirrand_current_list->is_randomized_l = 0;
		temp_id = dirrand_current_list->next_sequential;
	}
	dirrand_new_list(((playlist_t *)Addr6(current_file))->playlist, 1);
}



static simm_id
dirrand_new_list(simm_id /* playlist_list_t */ list_id, unsigned char use_current_file)
{
	playlist_t *entry;
	//unsigned int num_files=0;

	addr6_backup = addr6_val;
	addr7_backup = addr7_val;

	dirrand_current_list = (playlist_list_t *)Addr6(list_id);

	// we're already been to this dir, so no action needed
	if (dirrand_current_list->is_randomized_l) {
		return dirrand_current_list->playlist_start_random;
	}

	// a new dir we're not seen before... must initialize it
	temp_id = dirrand_current_list->playlist_start_sequential;
	dirrand_current_list->num_unrand_files = 0;
	dirrand_current_list->is_randomized_l = 1;
	while (temp_id) {
		entry = (playlist_t *)addr7(temp_id);
		entry->is_randomized = 0;
		entry->prev_random = 0;
		entry->next_random = 0;
		temp_id = entry->next_sequential;
		dirrand_current_list->num_unrand_files++;
	}

	if (dirrand_current_list->num_unrand_files > 1) {
		dirrand_current_list->num_unrand_files--;
		if (use_current_file) {
			entry = Addr7(current_file);
		} else {
			entry = dirrand_choose_file();
		}
		entry->is_randomized = 1;
		temp_id = simm_id_from_pointer(entry);
		dirrand_current_list->playlist_start_random = temp_id;
		dirrand_current_list->playlist_random_head = temp_id;
		dirrand_current_list->playlist_random_tail = temp_id;
		#ifdef VERBOSE_DEBUG
		print("Init rand one list, 1st=");
		print_hex32(temp_id);
		print_crlf();
		#endif
	} else {
		temp_id = dirrand_current_list->playlist_start_sequential;
		dirrand_current_list->playlist_start_random = temp_id;
		dirrand_current_list->num_unrand_files = 0;
		entry = (playlist_t *)addr7(temp_id);
		entry->prev_random = temp_id;
		entry->next_random = temp_id;
		entry->is_randomized = 1;
	}
	addr6_val = addr6_backup;
	addr7_val = addr7_backup;
	return temp_id;
}



static void
dirrand_randomize_incremental(simm_id /* playlist_t */ current)
{
	playlist_t *entry;

	addr6_backup = addr6_val;
	addr7_backup = addr7_val;

	entry = (playlist_t *)Addr7(current);
	dirrand_current_list = (playlist_list_t *)Addr6(entry->playlist);

	if (entry->next_random == 0) {
		entry = dirrand_choose_file();
		entry->is_randomized = 1;
		entry->prev_random = current;
		temp_id = simm_id_from_pointer(entry);
		entry = (playlist_t *)Addr7(current);
		entry->next_random = temp_id;
		dirrand_current_list->playlist_random_head = temp_id;
		dirrand_current_list->num_unrand_files--;
		#ifdef VERBOSE_DEBUG
		print("current=");
		print_hex32(current);
		print(" next=");
		print_hex32(temp_id);
		print_crlf();
		#endif
	}
	if (entry->prev_random == 0) {
		entry = dirrand_choose_file();
		entry->is_randomized = 1;
		entry->next_random = current;
		temp_id = simm_id_from_pointer(entry);
		entry = (playlist_t *)Addr7(current);
		entry->prev_random = temp_id;
		dirrand_current_list->playlist_random_tail = temp_id;
		dirrand_current_list->num_unrand_files--;
		#ifdef VERBOSE_DEBUG
		print("current=");
		print_hex32(current);
		print(" prev=");
		print_hex32(temp_id);
		print_crlf();
		#endif
	}
	if (dirrand_current_list->num_unrand_files == 0) {
		#ifdef VERBOSE_DEBUG
		print("Finalizing, head=");
		print_hex32(dirrand_current_list->playlist_random_head);
		print("tail=");
		print_hex32(dirrand_current_list->playlist_random_tail);
		print_crlf();
		#endif
		((playlist_t *)Addr7(dirrand_current_list->playlist_random_head))->
		  next_random = dirrand_current_list->playlist_random_tail;
		((playlist_t *)Addr7(dirrand_current_list->playlist_random_tail))->
		  prev_random = dirrand_current_list->playlist_random_head;
	}
	addr6_val = addr6_backup;
	addr7_val = addr7_backup;
}



/*************************************************************************************/
/* Randomization code for MODE_PLSTRAND (seq within each dir, choose dirs randomly)  */
/*************************************************************************************/


static playlist_list_t *
plstrand_choose_list(void)
{
	unsigned int count;
	playlist_list_t *list;

	list = (playlist_list_t *)Addr6(first_playlist);
	count = rand16() % num_unrand_lists;
	#ifdef VERBOSE_DEBUG
	printf("begin plstrand_choose_list, count=%u, num_unrand=%u\r\n",
		count, num_unrand_lists);
	#endif

	while (1) {
		if (!list->is_randomized_l) {
			#ifdef VERBOSE_DEBUG
			printf(" [%lx]", simm_id_from_pointer(list));
			#endif
			if (count == 0) {
				#ifdef VERBOSE_DEBUG
				print("\r\nend plstrand_choose_list\r\n");
				#endif
				return list;
			}
			count--;
		} else {
			#ifdef VERBOSE_DEBUG
			printf(" (%lx)", simm_id_from_pointer(list));
			#endif
		}
		list = (playlist_list_t *)Addr6(list->next_sequential);
	}
}




static void
plstrand_initialize(void)
{
	playlist_list_t *list;

	addr6_backup = addr6_val;

	/* mark all playlists as not randomized */
	num_unrand_lists = 0;
	temp_id = first_playlist;
	while (temp_id) {
		list = (playlist_list_t *)addr6(temp_id);
		list->is_randomized_l = 0;
		list->next_random = 0;
		list->prev_random = 0;
		temp_id = list->next_sequential;
		num_unrand_lists++;
	}
	if (num_unrand_lists > 1) {
		num_unrand_lists--;
		list = Addr6(((playlist_t *)Addr6(current_file))->playlist);
		list->is_randomized_l = 1;
		temp_id = simm_id_from_pointer(list);
		start_rand_playlist = temp_id;
		head_rand_playlist = temp_id;
		tail_rand_playlist = temp_id;
		#ifdef VERBOSE_DEBUG
        	print("Init rand lists, 1st=");
		print_hex32(temp_id);
		print_crlf();
		#endif
	} else {
		num_unrand_lists = 0;
		start_rand_playlist = first_playlist;
		list = (playlist_list_t *)Addr6(first_playlist);
		list->next_random = first_playlist;
		list->prev_random = first_playlist;
		list->is_randomized_l = 1;
	}
	addr6_val = addr6_backup;
}


static void
plstrand_randomize_incremental(simm_id /* playlist_list_t */ list_id)
{
	playlist_list_t *list;

	addr6_backup = addr6_val;
	list = (playlist_list_t *)Addr6(list_id);

	if (list->next_random == 0) {
		list = plstrand_choose_list();
		list->is_randomized_l = 1;
		list->prev_random = list_id;
		temp_id = simm_id_from_pointer(list);
		list = (playlist_list_t *)Addr6(list_id);
		list->next_random = temp_id;
		head_rand_playlist = temp_id;
		num_unrand_lists--;
		#ifdef VERBOSE_DEBUG
		print("list_id=");
		print_hex32(list_id);
		print(" next=");
		print_hex32(temp_id);
		print_crlf();
		#endif
	}
	if (list->prev_random == 0) {
		list = plstrand_choose_list();
		list->is_randomized_l = 1;
		list->next_random = list_id;
		temp_id = simm_id_from_pointer(list);
		list = (playlist_list_t *)Addr6(list_id);
		list->prev_random = temp_id;
		tail_rand_playlist = temp_id;
		num_unrand_lists--;
		#ifdef VERBOSE_DEBUG
		print("list_id=");
		print_hex32(list_id);
		print(" prev=");
		print_hex32(temp_id);
		print_crlf();
		#endif
	}
	if (num_unrand_lists == 0) {
		#ifdef VERBOSE_DEBUG
		print("Finalizing, head=");
		print_hex32(head_rand_playlist);
		print("tail=");
		print_hex32(tail_rand_playlist);
		print_crlf();
		#endif
		((playlist_list_t *)Addr6(head_rand_playlist))->next_random =
			tail_rand_playlist;
		((playlist_list_t *)Addr6(tail_rand_playlist))->prev_random =
			head_rand_playlist;
	}
	addr6_val = addr6_backup;
}








/*************************************************************************************/
/* Randomization code for MODE_ALLRAND (randomly choose files among all dirs)        */
/*************************************************************************************/


static playlist_t *
allrand_choose_file(void)
{
	unsigned int count;
	playlist_t *file;
	playlist_list_t *list;

	count = rand16() % num_unrand_all_files;
	#ifdef VERBOSE_DEBUG
	printf("begin allrand_choose_file, count=%u, num_unrand=%u\r\n",
		count, num_unrand_all_files);
	#endif

	list = (playlist_list_t *)Addr6(first_playlist);
	while (count >= list->num_unrand_files) {
		 #ifdef VERBOSE_DEBUG
		 print("skipping list ");
		 print_hex32(simm_id_from_pointer(list));
		 printf(" with %u files\r\n", list->num_unrand_files);
		 #endif
		count -= list->num_unrand_files;
		list = (playlist_list_t *)Addr6(list->next_sequential);
	}
	#ifdef VERBOSE_DEBUG
	print("using list ");
	print_hex32(simm_id_from_pointer(list));
	printf(" with %u files\r\n", list->num_unrand_files);
	#endif

	list->num_unrand_files--;

	file = (playlist_t *)Addr7(list->playlist_start_sequential);
	while (1) {
		if (!file->is_randomized) {
			#ifdef VERBOSE_DEBUG
			printf(" [%lx]", simm_id_from_pointer(file));
			#endif
			if (count == 0) {
				num_unrand_all_files--;
				#ifdef VERBOSE_DEBUG
				print("\r\nend allrand_choose_file\r\n");
				#endif
				return file;
			}
			count--;
		} else {
			#ifdef VERBOSE_DEBUG
			printf(" (%lx)", simm_id_from_pointer(file));
			#endif
		}
		// TODO: using Addr7 causes crash... is this a FPGA
		// timing problem, and why does addr7 not crash??
		// TODO: is this still a problem ???  Test with different simms
		file = (playlist_t *)Addr7(file->next_sequential);
	}
}


static unsigned int
allrand_init_one_list(simm_id list_id)
{
	unsigned int count=0;
	playlist_t *file;

	temp_id = list_id;
	while (temp_id) {
		file = (playlist_t *)addr7(temp_id);
		file->is_randomized = 0;
		file->prev_random = 0;
		file->next_random = 0;
		temp_id = file->next_sequential;
		count++;
	}
	return count;
}


static void
allrand_initialize(void)
{
	playlist_list_t *list;
	playlist_t *file;

	addr6_backup = addr6_val;
	addr7_backup = addr7_val;

	list = (playlist_list_t *)Addr6(first_playlist);
	num_unrand_all_files = 0;
	while (1) {
		num_unrand_all_files += list->num_unrand_files = 
		  allrand_init_one_list(list->playlist_start_sequential);

		if (list->next_sequential) {
			list = (playlist_list_t *)Addr6(list->next_sequential);
		} else {
			break;
		}
	}

	file = Addr7(current_file);
	file->is_randomized = 1;
	((playlist_list_t *)Addr6(file->playlist))->num_unrand_files--;
	num_unrand_all_files--;
	start_all_rand = current_file;
	head_all_rand = current_file;
	tail_all_rand = current_file;
	#ifdef VERBOSE_DEBUG
	print("Init rand all, 1st=");
	print_hex32(current_file);
	print_crlf();
	#endif
	addr6_val = addr6_backup;
	addr7_val = addr7_backup;
}



static void
allrand_randomize_incremental(simm_id /* playlist_t */ current)
{
	playlist_t *file;

	addr6_backup = addr6_val;
	addr7_backup = addr7_val;

	file = (playlist_t *)Addr7(current);

	if (file->next_random == 0) {
		file = allrand_choose_file();
		file->is_randomized = 1;
		file->prev_random = current;
		temp_id = simm_id_from_pointer(file);
		file = (playlist_t *)Addr7(current);
		file->next_random = temp_id;
		head_all_rand = temp_id;
		#ifdef VERBOSE_DEBUG
                print("current=");
		print_hex32(current);
		print(" next=");
		print_hex32(temp_id);
		print_crlf();
		#endif
	}
	if (file->prev_random == 0) {
		file = allrand_choose_file();
		file->is_randomized = 1;
		file->next_random = current;
		temp_id = simm_id_from_pointer(file);
		file = (playlist_t *)Addr7(current);
		file->prev_random = temp_id;
		tail_all_rand = temp_id;
		#ifdef VERBOSE_DEBUG
                print("current=");
		print_hex32(current);
		print(" prev=");
		print_hex32(temp_id);
		print_crlf();
		#endif
	}
	if (num_unrand_all_files == 0) {
		#ifdef VERBOSE_DEBUG
		print("Finalizing, head=");
		print_hex32(head_all_rand);
		print("tail=");
		print_hex32(tail_all_rand);
		print_crlf();
		#endif
		((playlist_t *)Addr7(head_all_rand))->next_random = tail_all_rand;
		((playlist_t *)Addr7(tail_all_rand))->prev_random = head_all_rand;
	}
	addr6_val = addr6_backup;
	addr7_val = addr7_backup;
}		








